import subprocess
from .vm import VM, Disk, NIC
from .config import Config
from cmdtool import SuperCommand, SubCommand, ToList


class BKeeper(SuperCommand):
    def __init__(self):
        super().__init__(name='bkeeper',
                         description='A tool for managing bhyve VM\'s',
                         subscripts=[CreateOnce, DestroyOnce, Create, Destroy, CreateAll,
                                     Add, Remove, Clone, Snapshot],
                         log_fmt='bkeeper: %(levelname)s: %(message)s')


class ConfigOps(SubCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_arg('--config', default='/usr/local/etc/bhyve.yaml')

    def load_config(self):
        return Config.open(self.args.config)

    def save(self, config):
        if not self.testmode:
            config.save()
        else:
            self.debug(config.dump())


class VMOps(ConfigOps):

    def load_vm(self, name):
        return self.load_config().get(name)

    def create_vm(self, name):
        self.info('creating VM: '+name)
        vm = self.load_vm(name)
        for command in vm.create_nics() + vm.start_bootloader():
            self.sh(command)
        try:
            self.sh(vm.start_os())
        except subprocess.CalledProcessError as e:
            if not e.output.decode() == '' and e.returncode == 1:
                raise
            else:
                self.debug('VM stopped with exception: ' + str(e))
        finally:
            self.destroy_vm(name)

    def destroy_vm(self, name):
        self.info('destroying VM: '+name)
        for command in self.load_vm(name).destroy():
            self.sh(command)


class Create(VMOps):
    def __init__(self, supercmd):
        super().__init__('create', supercmd)
        self.add_arg('name')

    def script(self):
        self.create_vm(self.args.name)


class Destroy(VMOps):
    def __init__(self, supercmd):
        super().__init__('destroy', supercmd)
        self.add_arg('name')

    def script(self):
        self.destroy_vm(self.args.name)


class CreateAll(VMOps):
    def __init__(self, supercmd):
        super().__init__('create_all', supercmd)

    def script(self):
        for name in self.load_config().vms.keys():
            self.run_thread(self.create_vm, name)


class VMCreation(SubCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_arg('name')
        self.add_arg('nmdm_id', type=int)
        self.add_arg('-c', '--cpus', type=int, default=1)
        self.add_arg('-m', '--memsize', default='256M')
        self.add_arg('-g', '--grubdir', default='/boot/grub')
        self.add_arg('-b', '--bootpart', default='gpt1')
        self.add_arg('-n', '--nic', dest='nics', action=ToList, nargs=2, required=True)
        self.add_arg('-d', '--disk', dest='disks', action=ToList, nargs=2, required=True)

    def get_vm(self):
        vm = {
            'name': self.args.name,
            'nmdm_id': self.args.nmdm_id,
            'cpus': self.args.cpus,
            'memsize': self.args.memsize,
            'grubdir': self.args.grubdir,
            'bootpart': self.args.bootpart,
            'nics': list(map(lambda x: NIC(*x), self.args.nics)),
            'disks': list(map(lambda x: Disk(*x), self.args.disks))
        }
        return VM(**vm)


class CreateOnce(VMCreation):
    def __init__(self, supercmd):
        super().__init__('create_once', supercmd)

    def script(self):
        vm = self.get_vm()
        self.info('creating VM: '+vm.name)
        for command in vm.create():
            self.sh(command)


class DestroyOnce(SubCommand):
    def __init__(self, supercmd):
        super().__init__('destroy_once', supercmd)
        self.add_arg('name')
        self.add_arg('-n', '--nic', dest='nics', action=ToList)

    def script(self):
        commands = [VM.destroy_once(self.args.name)]
        commands.extend(map(lambda x: NIC.destroy_once(x), self.args.nics))
        for command in commands:
            self.sh(command)


class Add(ConfigOps, VMCreation):
    def __init__(self, supercmd):
        super().__init__('add', supercmd)

    def script(self):
        vm = self.get_vm()
        config = self.load_config()
        config.add(vm)
        self.save(config)


class Remove(VMOps):
    def __init__(self, supercmd):
        super().__init__('remove', supercmd)
        self.add_arg('name')
        self.add_arg('--erase', action='store_true')

    def script(self):
        config = self.load_config()
        config.remove(self.args.name)
        self.save(config)

        if self.args.erase:
            vm = self.load_vm(self.args.name)
            for disk in vm.disks:
                self.sh(disk.destroy())


class Clone(ConfigOps):
    def __init__(self, supercmd):
        super().__init__('clone', supercmd)
        self.add_arg('source')
        self.add_arg('name')

    def script(self):
        config = self.load_config()
        for cmd in config.clone(self.args.source, self.args.name):
            self.sh(cmd)
        self.save(config)


class Snapshot(ConfigOps):
    def __init__(self, supercmd):
        super().__init__('snapshot', supercmd)
        self.add_arg('name')
        self.add_arg('snapshot')
        self.add_arg('-d', '--disk')

    def script(self):
        config = self.load_config()
        vm = config.get(self.args.name)
        if not self.args.disk:
            for cmd in [disk.snapshot(self.args.snapshot) for disk in vm.disks]:
                self.sh(cmd)
        else:
            try:
                disk = [d for d in vm.disks if d.name == self.args.disk][0]
            except IndexError:
                self.error('no disk with name: ' + self.args.disk)
            else:
                self.sh(disk.snapshot(self.args.snapshot))


import os
import copy
from .utils import flatmap
from .serializable import Serializable
from .vm import VM


class Config(Serializable):
    def __init__(self, vms):
        self.vms = vms
        self.file = ''

    def add(self, vm):
        self.vms[vm.name] = vm
        return self

    def remove(self, vm_name):
        del self.vms[vm_name]
        return self

    def modify(self, vm):
        return self.add(vm)

    def get(self, vm_name):
        return self.vms[vm_name]

    def clone(self, source, name):
        vm = copy.deepcopy(self.get(source))

        vm.name = name
        vm.nmdm_id = self.new_nmdmid()
        tapid = self.new_tapid()

        for nic in vm.nics:
            nic.name = 'tap' + str(tapid)
            tapid += 1

        cmds = []
        for disk in vm.disks:
            new_name = disk.name.replace(source, name)
            for cmd in disk.clone(new_name):
                cmds.append(cmd)
            disk.name = new_name

        self.add(vm)
        return cmds

    def to_dict(self):
        dct = {}
        for k, v in self.vms.items():
            dct[k] = v.to_dict()
            del dct[k]['name']
        return dct

    @classmethod
    def from_dict(cls, dct):
        if dct is None:
            dct = {}
        vms = {}
        for k, v in dct.items():
            v['name'] = k
            vms[k] = VM.from_dict(v)
        return cls(vms)

    @classmethod
    def open(cls, config_file):
        if os.path.exists(config_file):
            with open(config_file) as cf:
                config = cls.load(cf.read())
        else:
            config = cls.from_dict({})

        config.file = config_file
        return config

    def save(self):
        assert self.file
        with open(self.file, 'w') as cf:
            cf.write(self.dump())

    def new_tapid(self):
        max_id = max(map(lambda x: int(x.name[3:]), flatmap(lambda x: x.nics, self.vms.values())))
        return max_id + 1

    def new_nmdmid(self):
        return max(vm.nmdm_id for vm in self.vms.values()) + 1


from .utils import flatmap
from .serializable import Serializable, load_from_key_value


class VM(Serializable):
    def __init__(self, name, memsize, cpus, nmdm_id, bootpart, grubdir, nics, disks):
        self.name = name
        self.memsize = memsize
        self.cpus = cpus
        self.nmdm_id = nmdm_id
        self.bootpart = bootpart
        self.grubdir = grubdir
        self.nics = nics
        self.disks = disks

        self.nmdm = '/dev/nmdm{}A'.format(self.nmdm_id)
        self.devmap = '(hd0) ' + self.disks[0].zvol
        self.bootdev = 'hd0,' + self.bootpart

    def create_nics(self):
        return list(flatmap(lambda x: x.create(), self.nics))

    def start_bootloader(self):
        return ['echo "{devmap}" > /tmp/{name}-device.map'.format(**vars(self)),
                'grub-bhyve -m /tmp/{name}-device.map -d {grubdir}'
                ' -r {bootdev} -M {memsize} {name}'.format(**vars(self)),
                'rm /tmp/{name}-device.map'.format(**vars(self))]

    def start_os(self):
        options = [
            '-A', '-I', '-H', '-P',
            '-c ' + str(self.cpus),
            '-m ' + self.memsize,
            '-l com1,' + self.nmdm,
            '-s 0,hostbridge',
            '-s 1,lpc'
        ]
        for idx, item in enumerate(self.disks + self.nics):
            options.append(item.as_option(idx + 2))

        return 'bhyve {0} {1}'.format(' '.join(options), self.name)

    def create(self):
        return self.create_nics() + self.start_bootloader() + [self.start_os()]

    def destroy(self):
        return ['bhyvectl --destroy --vm='+self.name] + list(map(lambda x: x.destroy(), self.nics))

    @staticmethod
    def destroy_once(name):
        return 'bhyvectl --destroy --vm='+name

    def to_dict(self):
        my_vars = dict(vars(self))
        del my_vars['nmdm']
        del my_vars['devmap']
        del my_vars['bootdev']
        my_vars['nics'] = list(map(lambda x: x.to_dict(), self.nics))
        my_vars['disks'] = list(map(lambda x: x.to_dict(), self.disks))
        return my_vars

    @classmethod
    def from_dict(cls, dct):
        dct['nics'] = list(map(lambda x: NIC.from_dict(x), dct['nics']))
        dct['disks'] = list(map(lambda x: Disk.from_dict(x), dct['disks']))
        return super().from_dict(dct)


class NIC(Serializable):
    def __init__(self, name, bridge):
        self.name = name
        self.bridge = bridge

    def create(self):
        return ['ifconfig {} create'.format(self.name),
                'ifconfig {0} addm {1}'.format(self.bridge, self.name)]

    def destroy(self):
        return 'ifconfig {} destroy'.format(self.name)

    @staticmethod
    def destroy_once(name):
        return 'ifconfig {} destroy'.format(name)

    def as_option(self, slot):
        return '-s {0},virtio-net,{1}'.format(slot, self.name)

    def to_dict(self):
        return {self.name: self.bridge}

    @classmethod
    def from_dict(cls, dct):
        return load_from_key_value(cls, dct)


class Disk(Serializable):
    def __init__(self, name, pool, size=None):
        self.name = name
        self.pool = pool
        self.size = size

        self.zvol = '/dev/zvol/{0}/{1}'.format(self.pool, self.name)

    def snapshot(self, name):
        return 'zfs snapshot {0}/{1}@{2}'.format(self.pool, self.name, name)

    def clone(self, name):
        return [self.snapshot(name), 'zfs clone {0}/{1}@{2} {0}/{2}'.format(self.pool, self.name, name)]

    def create(self):
        assert self.size
        return 'zfs create -V {size} {pool}/{name}'.format(**vars(self))

    def destroy(self):
        return 'zfs destroy {pool}/{name}'.format(**vars(self))

    def as_option(self, slot):
        return '-s {0},virtio-blk,{1}'.format(slot, self.zvol)

    def to_dict(self):
        return {self.name: self.pool}

    @classmethod
    def from_dict(cls, dct):
        return load_from_key_value(cls, dct)


import yaml


class Serializable:
    def to_dict(self):
        vars(self)

    def dump(self):
        return yaml.dump(self.to_dict(), default_flow_style=False)

    @classmethod
    def from_dict(cls, dct):
        return cls(**dct)

    @classmethod
    def load(cls, yaml_str):
        return cls.from_dict(yaml.load(yaml_str))


def load_from_key_value(cls, dct):
    k, v = list(dct.items())[0]
    return cls(k, v)



#!/usr/bin/env python3
from setuptools import setup

setup(name='bhyve',
      version='0.2.1',
      description='A tool for managing bhyve VM\'s, '
                  'because libvirt doesn\'t do want I want it to',
      url='https://github.com/ConnorDillon/bhyve',
      author='Connor Dillon',
      author_email='connor@cdillon.nl',
      license='GPLv3',
      packages=['bhyve'],
      scripts=['scripts/bkeep'],
      install_requires=['PyYAML', 'cmdtool==0.2'],
      test_suite='bhyve.tests',
      zip_safe=False)
